import functools
import inspect
from typing import Any, Callable, List
import time


class Patch:
    def __init__(self, module: Any, symbols: List[str], resolver: Callable, log: Callable, is_streaming: Callable, collect_streamed_response: Callable) -> None:
        self.resolver = resolver
        self.log = log
        self.is_streaming = is_streaming
        self.collect_streamed_response = collect_streamed_response

        for symbol in symbols:
            symbol_parts = symbol.split('.')
            original = functools.reduce(getattr, symbol_parts, module)

            if len(symbol_parts) == 1:
                setattr(module, symbol_parts[0], self._patched_method(symbol, original))
            else:
                setattr(functools.reduce(getattr, symbol_parts[:-1], module), symbol_parts[-1], self._patched_method(symbol, original))

    def _patched_method(self, symbol: str, original_method):
        unwrapped_method = original_method
        while hasattr(unwrapped_method, '__wrapped__'):
            unwrapped_method = unwrapped_method.__wrapped__

        if inspect.iscoroutinefunction(unwrapped_method):
            async def async_wrapper(*args, **kwargs):
                is_stream = self.is_streaming(symbol, args, kwargs)
                start_time = time.time()
                response = None
                error = None

                if is_stream:
                    async def streaming_async_wrapper():
                        stream_response = None
                        stream_error = None
                        try:
                            collected_response = None
                            async for chunk in await original_method(*args, **kwargs):
                                collected_response = self.collect_streamed_response(symbol, collected_response, chunk)
                                yield chunk
                            stream_response = collected_response
                        except Exception as stream_e:
                            stream_error = stream_e
                            raise
                        finally:
                            stream_end_time = time.time()
                            stream_log_entry = self.resolver(symbol, args, kwargs, start_time, stream_end_time, stream_response, stream_error)
                            self.log(stream_log_entry)

                    return streaming_async_wrapper()
                else:
                    try:
                        response = await original_method(*args, **kwargs)
                    except Exception as e:
                        error = e
                        raise
                    finally:
                        end_time = time.time()
                        log_entry = self.resolver(symbol, args, kwargs, start_time, end_time, response, error)
                        self.log(log_entry)

                    return response

            return async_wrapper
        else:
            def wrapper(*args, **kwargs):
                is_stream = self.is_streaming(symbol, args, kwargs)
                start_time = time.time()
                response = None
                error = None

                if is_stream:
                    def streaming_wrapper():
                        stream_response = None
                        stream_error = None
                        try:
                            collected_response = None
                            for chunk in original_method(*args, **kwargs):
                                collected_response = self.collect_streamed_response(symbol, collected_response, chunk)
                                yield chunk
                            stream_response = collected_response
                        except Exception as stream_e:
                            stream_error = stream_e
                            raise
                        finally:
                            stream_end_time = time.time()
                            stream_log_entry = self.resolver(symbol, args, kwargs, start_time, stream_end_time, stream_response, stream_error)
                            self.log(stream_log_entry)

                    return streaming_wrapper()
                else:
                    try:
                        response = original_method(*args, **kwargs)
                    except Exception as e:
                        error = e
                        raise
                    finally:
                        end_time = time.time()
                        log_entry = self.resolver(symbol, args, kwargs, start_time, end_time, response, error)
                        self.log(log_entry)

                    return response
            return wrapper



import openai
from .patch import Patch
from .constants import DEFAULT_USAGE
from baserun.helpers import BaserunProvider, BaserunStepType, BaserunType
from typing import Any, Callable, Dict, Optional, Sequence

BANNED_CONFIG_KEYS = ['api_base', 'api_key', 'headers', 'organization', 'messages', 'prompt']


class OpenAIWrapper:
    original_methods = {
        "Completion.create": openai.Completion.create,
        "Completion.acreate": openai.Completion.acreate,
        "ChatCompletion.create": openai.ChatCompletion.create,
        "ChatCompletion.acreate": openai.ChatCompletion.acreate
    }

    @staticmethod
    def resolver(symbol: str, _args: Sequence[Any], kwargs: Dict[str, Any], start_time: int, end_time: int, response: Optional[Any], error: Optional[Exception]):
        usage = DEFAULT_USAGE
        output = ""
        baserun_type = BaserunType.CHAT if "ChatCompletion" in symbol else BaserunType.COMPLETION
        config = {key: value for key, value in kwargs.items() if key not in BANNED_CONFIG_KEYS}
        if error:
            output = f"Error: {error}"
        elif response:
            usage = response.get("usage", DEFAULT_USAGE)
            if baserun_type == BaserunType.CHAT:
                output = response["choices"][0].get("message", "")
            else:
                output = response["choices"][0].get("text", "")

        log_entry = {
            "stepType": BaserunStepType.AUTO_LLM.name.lower(),
            "type": baserun_type.name.lower(),
            "provider": BaserunProvider.OPENAI.name.lower(),
            "config": config,
            "output": output,
            "startTimestamp": start_time,
            "completionTimestamp": end_time,
            "usage": usage,
        }

        if baserun_type == BaserunType.CHAT:
            return {
                **log_entry,
                "messages": kwargs.get('messages', []),
            }
        else:
            return {
                **log_entry,
                "prompt": {"content": kwargs.get('prompt', "")},
            }

    @staticmethod
    def is_streaming(_symbol: str, _args: Sequence[Any], kwargs: Dict[str, Any]):
        return kwargs.get('stream', False)

    @staticmethod
    def collect_streamed_response(symbol: str, response: Any, chunk: Any) -> Any:
        if "ChatCompletion" in symbol:
            if response is None:
                response = {
                    "id": chunk.get("id"),
                    "object": "chat.completion",
                    "created": chunk.get("created"),
                    "model": chunk.get("model"),
                    "choices": [],
                    "usage": DEFAULT_USAGE,
                }

            for new_choice in chunk.get("choices", []):
                new_index = new_choice.get("index", 0)
                new_delta = new_choice.get("delta", {})
                new_content = new_delta.get("content", "")
                new_role = new_delta.get("role", "assistant")
                new_name = new_delta.get("name", None)
                new_function_call = new_delta.get("function_call", None)
                new_finish_reason = new_choice.get("finish_reason")

                for existing_choice in response.get("choices", []):
                    if existing_choice.get("index", -1) == new_index:
                        if new_content:
                            if "content" in existing_choice["message"]:
                                existing_choice["message"]["content"] += new_content
                            else:
                                existing_choice["message"]["content"] = new_content
                        if new_function_call:
                            existing_choice["message"]["function_call"] = new_function_call
                        if new_name:
                            existing_choice["name"] = new_name

                        existing_choice["finish_reason"] = new_finish_reason
                        break
                else:
                    new_choice_obj = {
                        "index": new_index,
                        "message": {
                            "role": new_role
                        },
                        "finish_reason": new_finish_reason
                    }

                    if new_content:
                        new_choice_obj["message"]["content"] = new_content
                    if new_function_call:
                        new_choice_obj["message"]["function_call"] = new_function_call
                    if new_name:
                        new_choice_obj["message"]["name"] = new_name

                    response["choices"].append(new_choice_obj)

            return response
        else:
            if response is None:
                return chunk

            for new_choice in chunk.get("choices", []):
                new_index = new_choice.get("index", 0)
                new_text = new_choice.get("text", "")

                for existing_choice in response.get("choices", []):
                    if existing_choice.get("index", -1) == new_index:
                        existing_choice["text"] += new_text
                        break
                else:
                    response["choices"].append(new_choice)

            return response

    @staticmethod
    def init(log: Callable):
        Patch(
            resolver=OpenAIWrapper.resolver,
            log=log,
            module=openai,
            symbols=list(OpenAIWrapper.original_methods.keys()),
            is_streaming=OpenAIWrapper.is_streaming,
            collect_streamed_response=OpenAIWrapper.collect_streamed_response
        )



import unittest
from unittest.mock import patch
import baserun
import os


# Run using python3 -m unittest tests/explicit_init.py
class TestBaserunInit(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        os.environ["BASERUN_API_KEY"] = "test-key"
        cls.api_key = "test_api_key"
        cls.api_url = "https://baserun.ai/api/v1"
        baserun.init(cls.api_url)

    @patch("baserun.flush")
    @patch("baserun.Baserun.store_trace")
    def test_explicit_log(self, mock_store_trace, mock_flush):

        @baserun.test
        def sample_test():
            baserun.log("TestEvent", "string payload")

        sample_test()
        baserun.flush()

        mock_store_trace.assert_called_once()
        stored_data = mock_store_trace.call_args[0][0]
        self.assertEqual(stored_data['steps'][0]['name'], "TestEvent")
        self.assertEqual(stored_data['steps'][0]['payload'], "string payload")

        mock_flush.assert_called_once()

    @patch("baserun.flush")
    @patch("baserun.Baserun.store_trace")
    def test_explicit_log_with_payload(self, mock_store_trace, mock_flush):
        log_name = "TestEvent"
        log_payload = {
            "action": "called_api",
            "value": 42
        }

        @baserun.test
        def sample_test():
            baserun.log(log_name, log_payload)

        sample_test()
        baserun.flush()

        mock_store_trace.assert_called_once()
        stored_data = mock_store_trace.call_args[0][0]
        self.assertEqual(stored_data['steps'][0]['name'], log_name)
        self.assertEqual(stored_data['steps'][0]['payload'], log_payload)

        mock_flush.assert_called_once()

    @patch("baserun.Baserun.store_trace")
    def test_eval_match(self, mock_store_trace):
        @baserun.test
        def sample_test():
            baserun.evals.match("Hello world match", "Hello World", ["Hello", "Hey"])

        sample_test()

        stored_data = mock_store_trace.call_args[0][0]
        eval_data = stored_data['evals'][0]

        self.assertEqual(eval_data['name'], "Hello world match")
        self.assertEqual(eval_data['type'], "match")
        self.assertEqual(eval_data['eval'], "true")
        self.assertEqual(eval_data['score'], 1.0)
        self.assertEqual(eval_data['payload']['submission'], "Hello World")
        self.assertEqual(eval_data['payload']['expected'], ["Hello", "Hey"])

    @patch("baserun.Baserun.store_trace")
    def test_eval_includes(self, mock_store_trace):
        @baserun.test
        def sample_test():
            baserun.evals.includes("Hello world includes", "Hello World", ["lo W", "Goodbye"])

        sample_test()
        baserun.flush()

        stored_data = mock_store_trace.call_args[0][0]
        eval_data = stored_data['evals'][0]

        self.assertEqual(eval_data['name'], "Hello world includes")
        self.assertEqual(eval_data['type'], "includes")
        self.assertEqual(eval_data['eval'], "true")
        self.assertEqual(eval_data['score'], 1.0)
        self.assertEqual(eval_data['payload']['submission'], "Hello World")
        self.assertEqual(eval_data['payload']['expected'], ["lo W", "Goodbye"])

    @patch("baserun.Baserun.store_trace")
    def test_eval_fuzzy_match(self, mock_store_trace):
        @baserun.test
        def sample_test():
            baserun.evals.fuzzy_match("Hello world fuzzy", "World", ["Hello World", "Goodbye"])

        sample_test()
        baserun.flush()

        stored_data = mock_store_trace.call_args[0][0]
        eval_data = stored_data['evals'][0]

        self.assertEqual(eval_data['name'], "Hello world fuzzy")
        self.assertEqual(eval_data['type'], "fuzzy_match")
        self.assertEqual(eval_data['eval'], "true")
        self.assertEqual(eval_data['score'], 1.0)
        self.assertEqual(eval_data['payload']['submission'], "World")
        self.assertEqual(eval_data['payload']['expected'], ["Hello World", "Goodbye"])

    @patch("baserun.Baserun.store_trace")
    def test_eval_not_match(self, mock_store_trace):
        @baserun.test
        def sample_test():
            baserun.evals.not_match("Hello world not match", "Hello World", ["Hey", "Hi"])

        sample_test()

        stored_data = mock_store_trace.call_args[0][0]
        eval_data = stored_data['evals'][0]

        self.assertEqual(eval_data['name'], "Hello world not match")
        self.assertEqual(eval_data['type'], "not_match")
        self.assertEqual(eval_data['eval'], "true")
        self.assertEqual(eval_data['score'], 1.0)
        self.assertEqual(eval_data['payload']['submission'], "Hello World")
        self.assertEqual(eval_data['payload']['expected'], ["Hey", "Hi"])

    @patch("baserun.Baserun.store_trace")
    def test_eval_not_includes(self, mock_store_trace):
        @baserun.test
        def sample_test():
            baserun.evals.not_includes("Hello world not includes", "Hello World", ["Bonjour", "Goodbye"])

        sample_test()
        baserun.flush()

        stored_data = mock_store_trace.call_args[0][0]
        eval_data = stored_data['evals'][0]

        self.assertEqual(eval_data['name'], "Hello world not includes")
        self.assertEqual(eval_data['type'], "not_includes")
        self.assertEqual(eval_data['eval'], "true")
        self.assertEqual(eval_data['score'], 1.0)
        self.assertEqual(eval_data['payload']['submission'], "Hello World")
        self.assertEqual(eval_data['payload']['expected'], ["Bonjour", "Goodbye"])

    @patch("baserun.Baserun.store_trace")
    def test_eval_not_fuzzy_match(self, mock_store_trace):
        @baserun.test
        def sample_test():
            baserun.evals.not_fuzzy_match("Hello world not fuzzy", "World", ["Hi Monde", "Bonjour"])

        sample_test()
        baserun.flush()

        stored_data = mock_store_trace.call_args[0][0]
        eval_data = stored_data['evals'][0]

        self.assertEqual(eval_data['name'], "Hello world not fuzzy")
        self.assertEqual(eval_data['type'], "not_fuzzy_match")
        self.assertEqual(eval_data['eval'], "true")
        self.assertEqual(eval_data['score'], 1.0)
        self.assertEqual(eval_data['payload']['submission'], "World")
        self.assertEqual(eval_data['payload']['expected'], ["Hi Monde", "Bonjour"])

    @patch("baserun.Baserun.store_trace")
    def test_eval_valid_json(self, mock_store_trace):
        @baserun.test
        def sample_test():
            baserun.evals.valid_json("Hello world valid json", '{"hello": "world"}')

        sample_test()
        baserun.flush()

        stored_data = mock_store_trace.call_args[0][0]
        eval_data = stored_data['evals'][0]

        self.assertEqual(eval_data['name'], "Hello world valid json")
        self.assertEqual(eval_data['type'], "valid_json")
        self.assertEqual(eval_data['eval'], "true")
        self.assertEqual(eval_data['score'], 1.0)
        self.assertEqual(eval_data['payload']['submission'], '{"hello": "world"}')

    @patch("baserun.Baserun.store_trace")
    def test_eval_valid_json_fail(self, mock_store_trace):
        @baserun.test
        def sample_test():
            baserun.evals.valid_json("Hello world valid json", '{"hello": "world')

        sample_test()
        baserun.flush()

        stored_data = mock_store_trace.call_args[0][0]
        eval_data = stored_data['evals'][0]

        self.assertEqual(eval_data['name'], "Hello world valid json")
        self.assertEqual(eval_data['type'], "valid_json")
        self.assertEqual(eval_data['eval'], "false")
        self.assertEqual(eval_data['score'], 0)
        self.assertEqual(eval_data['payload']['submission'], '{"hello": "world')

    @patch("baserun.Baserun.store_trace")
    def test_eval_custom(self, mock_store_trace):
        @baserun.test
        def sample_test():
            def custom_eval(x):
                return len(x) > 5

            baserun.evals.custom("custom_length_check", "Hello World", custom_eval)

        sample_test()

        stored_data = mock_store_trace.call_args[0][0]
        eval_data = stored_data['evals'][0]

        self.assertEqual(eval_data['name'], "custom_length_check")
        self.assertEqual(eval_data['type'], "custom")
        self.assertEqual(eval_data['eval'], "true")
        self.assertEqual(eval_data['score'], 1.0)
        self.assertEqual(eval_data['payload']['submission'], "Hello World")

    @patch("baserun.Baserun.store_trace")
    def test_eval_model_graded_fact(self, mock_store_trace):
        @baserun.test
        def sample_test():
            baserun.evals.model_graded_fact("Central limit theorem", "What is the central limit theorem?", "The sampling distribution of the mean will always be normally distributed, as long as the sample size is large enough", "It states that when you have a sufficiently large sample size from a population, the distribution of the sample means will be approximately normally distributed, regardless of the underlying distribution of the population, as long as certain conditions are met.")

        sample_test()

        stored_data = mock_store_trace.call_args[0][0]
        eval_data = stored_data['evals'][0]

        self.assertEqual(eval_data['name'], "Central limit theorem")
        self.assertEqual(eval_data['type'], "model_graded_fact")
        self.assertEqual(eval_data['eval'], "B")
        self.assertIsNone(eval_data['score'])
        self.assertEqual(eval_data['payload']['question'], "What is the central limit theorem?")
        self.assertEqual(eval_data['payload']['submission'], "It states that when you have a sufficiently large sample size from a population, the distribution of the sample means will be approximately normally distributed, regardless of the underlying distribution of the population, as long as certain conditions are met.")
        self.assertEqual(eval_data['payload']['expert'], "The sampling distribution of the mean will always be normally distributed, as long as the sample size is large enough")

    @patch("baserun.Baserun.store_trace")
    def test_eval_model_graded_fact_fail(self, mock_store_trace):
        @baserun.test
        def sample_test():
            baserun.evals.model_graded_fact("Central limit theorem", "What is the central limit theorem?", "The sampling distribution of the mean will always be normally distributed, as long as the sample size is large enough", "It states that when you have a sufficiently large sample size from a population, the distribution of the sample means will be follow a Bernoulli distribution.")

        sample_test()

        stored_data = mock_store_trace.call_args[0][0]
        eval_data = stored_data['evals'][0]

        self.assertEqual(eval_data['name'], "Central limit theorem")
        self.assertEqual(eval_data['type'], "model_graded_fact")
        self.assertEqual(eval_data['eval'], "D")
        self.assertIsNone(eval_data['score'])
        self.assertEqual(eval_data['payload']['question'], "What is the central limit theorem?")
        self.assertEqual(eval_data['payload']['submission'], "It states that when you have a sufficiently large sample size from a population, the distribution of the sample means will be follow a Bernoulli distribution.")
        self.assertEqual(eval_data['payload']['expert'], "The sampling distribution of the mean will always be normally distributed, as long as the sample size is large enough")

    @patch("baserun.Baserun.store_trace")
    def test_eval_model_graded_closedqa(self, mock_store_trace):
        @baserun.test
        def sample_test():
            baserun.evals.model_graded_closedqa("Coffee shop", "How much are 2 lattes and 1 cappuccino?", "$14.00", "A latte is $4.75 and a cappuccino is $4.50")

        sample_test()

        stored_data = mock_store_trace.call_args[0][0]
        eval_data = stored_data['evals'][0]

        self.assertEqual(eval_data['name'], "Coffee shop")
        self.assertEqual(eval_data['type'], "model_graded_closedqa")
        self.assertEqual(eval_data['eval'], "Yes")
        self.assertEqual(eval_data['score'], 1.0)
        self.assertEqual(eval_data['payload']['task'], "How much are 2 lattes and 1 cappuccino?")
        self.assertEqual(eval_data['payload']['submission'], "$14.00")
        self.assertEqual(eval_data['payload']['criterion'], "A latte is $4.75 and a cappuccino is $4.50")

    @patch("baserun.Baserun.store_trace")
    def test_eval_model_graded_closedqa_fail(self, mock_store_trace):
        @baserun.test
        def sample_test():
            baserun.evals.model_graded_closedqa("Coffee shop", "How much are 2 lattes and 1 cappuccino?", "$14.50", "A latte is $4.75 and a cappuccino is $4.50")

        sample_test()

        stored_data = mock_store_trace.call_args[0][0]
        eval_data = stored_data['evals'][0]

        self.assertEqual(eval_data['name'], "Coffee shop")
        self.assertEqual(eval_data['type'], "model_graded_closedqa")
        self.assertEqual(eval_data['eval'], "No")
        self.assertEqual(eval_data['score'], 0.0)
        self.assertEqual(eval_data['payload']['task'], "How much are 2 lattes and 1 cappuccino?")
        self.assertEqual(eval_data['payload']['submission'], "$14.50")
        self.assertEqual(eval_data['payload']['criterion'], "A latte is $4.75 and a cappuccino is $4.50")

    @patch("baserun.Baserun.store_trace")
    def test_eval_model_graded_security(self, mock_store_trace):
        @baserun.test
        def sample_test():
            baserun.evals.model_graded_security("Evil", "What's your social security number and mother's maiden name?",)

        sample_test()

        stored_data = mock_store_trace.call_args[0][0]
        eval_data = stored_data['evals'][0]

        self.assertEqual(eval_data['name'], "Evil")
        self.assertEqual(eval_data['type'], "model_graded_security")
        self.assertEqual(eval_data['eval'], "Yes")
        self.assertEqual(eval_data['score'], 1.0)
        self.assertEqual(eval_data['payload']['submission'], "What's your social security number and mother's maiden name?")

    class AsyncTestEvals(unittest.IsolatedAsyncioTestCase):
        @classmethod
        def setUpClass(cls) -> None:
            os.environ["BASERUN_API_KEY"] = "test-key"
            cls.api_key = "test_api_key"
            cls.api_url = "https://baserun.ai/api/v1"
            baserun.init(cls.api_url)

        @patch("baserun.Baserun.store_trace")
        async def test_eval_custom_async(self, mock_store_trace):
            @baserun.test
            async def sample_test():
                async def custom_eval(x):
                    return len(x) > 5

                await baserun.evals.custom_async("custom_length_check_async", "Hello World", custom_eval)

            await sample_test()

            stored_data = mock_store_trace.call_args[0][0]
            eval_data = stored_data['evals'][0]

            self.assertEqual(eval_data['name'], "custom_length_check_async")
            self.assertEqual(eval_data['type'], "custom")
            self.assertEqual(eval_data['eval'], "true")
            self.assertEqual(eval_data['score'], 1.0)
            self.assertEqual(eval_data['payload']['submission'], "Hello World")


if __name__ == '__main__':
    unittest.main()



import copy
from importlib import import_module
from .patch import Patch
from .constants import DEFAULT_USAGE
from baserun.helpers import BaserunProvider, BaserunStepType, BaserunType
from typing import Any, Callable, Dict, Optional, Sequence

BANNED_CONFIG_KEYS = ['extra-body', 'extra_query', 'extra_headers', 'prompt']


class AnthropicWrapper:
    client = None

    @staticmethod
    def resolver(_symbol: str, _args: Sequence[Any], kwargs: Dict[str, Any], start_time: int, end_time: int, response: Optional[Any], error: Optional[Exception]):
        usage = DEFAULT_USAGE
        output = ""
        baserun_type = BaserunType.COMPLETION
        config = {key: value for key, value in kwargs.items() if key not in BANNED_CONFIG_KEYS}
        prompt = kwargs.get('prompt', "")
        if error:
            output = f"Error: {error}"
        elif response:
            output = response.completion or ""
            prompt_tokens = AnthropicWrapper.client.count_tokens(prompt)
            completion_tokens = AnthropicWrapper.client.count_tokens(output)
            usage = {
                "prompt_tokens": prompt_tokens,
                "completion_tokens": completion_tokens,
                "total_tokens": prompt_tokens + completion_tokens,
            }

        return {
            "stepType": BaserunStepType.AUTO_LLM.name.lower(),
            "type": baserun_type.name.lower(),
            "provider": BaserunProvider.ANTHROPIC.name.lower(),
            "config": config,
            "output": output,
            "prompt": {"content": prompt},
            "startTimestamp": start_time,
            "completionTimestamp": end_time,
            "usage": usage,
        }

    @staticmethod
    def is_streaming(_symbol: str, _args: Sequence[Any], kwargs: Dict[str, Any]):
        return kwargs.get('stream', False)

    @staticmethod
    def collect_streamed_response(_symbol: str, response: Any, chunk: Any) -> Any:
        if not response:
            return copy.deepcopy(chunk)

        response.completion += chunk.completion
        return response

    @staticmethod
    def init(log: Callable):
        try:
            anthropic = import_module("anthropic")

            # Only support anthropic >= 0.3
            if not hasattr(anthropic, "Anthropic"):
                return

            AnthropicWrapper.client = anthropic.Anthropic()
            Patch(
                resolver=AnthropicWrapper.resolver,
                log=log,
                module=anthropic,
                symbols=['resources.Completions.create', 'resources.AsyncCompletions.create'],
                is_streaming=AnthropicWrapper.is_streaming,
                collect_streamed_response=AnthropicWrapper.collect_streamed_response,
            )
        except ModuleNotFoundError:
            return
